from typing import List, Optional, Tuple, Union

import re
import numpy as np
import pandas as pd
from pydantic import BaseModel, ValidationError

from classification_model.config.core import config
from classification_model.processing.data_manager import pre_processing




def validate_inputs(*, input_data: pd.DataFrame) -> Tuple[pd.DataFrame, Optional[dict]]:
    """Check model inputs for unprocessable values."""
    #pre_processed = pre_processing(dataframe=input_data)
    validated_data = input_data[config.model_config.features].copy()
    errors = None

    try:
        # replace numpy nans so that pydantic can validate
        MultipleTitanicDataInputs(
            inputs=validated_data.replace({np.nan: None}).to_dict(orient="records")
        )
    except ValidationError as error:
        errors = error.json()

    return validated_data, errors


class TitanicDataInputSchema(BaseModel):
    pclass: Optional[int]
    name: Optional[str]
    sex: Optional[str]
    age: Optional[int]
    sibsp: Optional[int]
    parch: Optional[int] 
    ticket: Optional[int]
    fare : Optional[float]
    cabin: Optional[str]
    embarked: Optional[str]
    title: Optional[str]
    boat: Optional[Union[str, int]]
    body: Optional[int]
    #home.dest: 

class MultipleTitanicDataInputs(BaseModel):
    inputs: List[TitanicDataInputSchema]





from typing import List

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
import re


class ExtractLetterTransformer(BaseEstimator, TransformerMixin):
    # Extract fist letter of variable

    def __init__(self, variable: str):
        self.variable = variable

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        # we need this step to fit the sklearn pipeline
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        # so that we do not over-write the original dataframe
        X = X.copy()
        
        X[self.variable] = X[self.variable].str[0]

        return X

class RareLabelCategoricalEncoder(BaseEstimator, TransformerMixin):
    """Groups infrequent categories into a single string"""

    def __init__(self, variables, rare_perc):

        if not isinstance(variables, list):
            raise ValueError('variables should be a list')
        
        self.rare_perc = rare_perc
        self.variables = variables

    def fit(self, X:pd.DataFrame, y=None):
        # persist frequent labels in dictionary
        self.encoder_dict_ = {}

        for var in self.variables:
            # the encoder will learn the most frequent categories
            t = pd.Series(X[var].value_counts(normalize=True)) 
            # frequent labels:
            self.encoder_dict_[var] = list(t[t >= self.rare_perc].index)

        return self

    def transform(self, X):
        X = X.copy()
        for feature in self.variables:
            X[feature] = np.where(
                X[feature].isin(self.encoder_dict_[feature]),
                                X[feature], "Rare")
        return X


import logging
import typing as t
from typing import Any, Union
from pathlib import Path

import joblib
import pandas as pd
from sklearn.pipeline import Pipeline
import numpy as np
from classification_model import __version__ as _version
from classification_model.config.core import DATASET_DIR, TRAINED_MODEL_DIR, config
import re


logger = logging.getLogger(__name__)


# retain only the first cabin if more than
# 1 are available per passenger
# float type for np.nan
def get_first_cabin(row: Any) -> Union[str, float]:
    try:
        return row.split()[0]
    except AttributeError:
        return np.nan

# extracts the title (Mr, Ms, etc) from the name variable
def get_title(passenger: str) -> str:
    line = passenger
    if re.search('Mrs', line):
        return 'Mrs'
    elif re.search('Mr', line):
        return 'Mr'
    elif re.search('Miss', line):
        return 'Miss'
    elif re.search('Master', line):
        return 'Master'
    else:
        return 'Other'

def pre_processing(*, dataframe:pd.DataFrame) -> pd.DataFrame:
    # replace interrogation marks by NaN values
    dataframe = dataframe.replace('?', np.nan)

    # retain only the first cabin if more than
    dataframe[config.model_config.CABIN] = dataframe[config.model_config.CABIN].apply(get_first_cabin)

    # extracts the title (Mr, Ms, etc) from the name variable
    dataframe['title'] = dataframe['name'].apply(get_title)

    # cast numerical variables as floats
    dataframe['fare'] = dataframe['fare'].astype('float')
    dataframe['age'] = dataframe['age'].astype('float')

    # drop unnecessary variables
    dataframe.drop(labels=config.model_config.unused_fields, axis=1, inplace=True)

    return dataframe


def _load_raw_dataset(*, file_name: str) -> pd.DataFrame:
    dataframe = pd.read_csv(Path(f"{DATASET_DIR}/{file_name}"))
    return dataframe

def load_dataset(*, file_name: str) -> pd.DataFrame:
    input_data = pd.read_csv(Path(f"{DATASET_DIR}/{file_name}"))
    transformed = pre_processing(dataframe=input_data)
    return transformed


def save_pipeline(*, pipeline_to_persist: Pipeline) -> None:
    """Persist the pipeline.
    Saves the versioned model, and overwrites any previous
    saved models. This ensures that when the package is
    published, there is only one trained model that can be
    called, and we know exactly how it was built.
    """

    # Prepare versioned save file name
    save_file_name = f"{config.app_config.pipeline_save_file}{_version}.pkl"
    save_path = TRAINED_MODEL_DIR / save_file_name

    remove_old_pipelines(files_to_keep=[save_file_name])
    joblib.dump(pipeline_to_persist, save_path)


def load_pipeline(*, file_name: str) -> Pipeline:
    """Load a persisted pipeline."""

    file_path = TRAINED_MODEL_DIR / file_name
    trained_model = joblib.load(filename=file_path)
    return trained_model


def remove_old_pipelines(*, files_to_keep: t.List[str]) -> None:
    """
    Remove old model pipelines.
    This is to ensure there is a simple one-to-one
    mapping between the package version and the model
    version to be imported and used by other applications.
    """
    do_not_delete = files_to_keep + ["__init__.py"]
    for model_file in TRAINED_MODEL_DIR.iterdir():
        if model_file.name not in do_not_delete:
            model_file.unlink()



from classification_model.config.core import config
from classification_model.processing.features import ExtractLetterTransformer


def test_extractletter_variable_transformer(sample_input_data):
    # Given
    #print(sample_input_data.loc[10,'cabin'])
    transformer = ExtractLetterTransformer(
        variable=config.model_config.CABIN,  # cabin
    )
    assert sample_input_data[config.model_config.CABIN].iat[10] == 'A6'

    # When
    subject = transformer.fit_transform(sample_input_data)

    # Then
    assert subject[config.model_config.CABIN].iat[10] == 'A'



import math

import numpy as np

from classification_model.predict import make_prediction
from sklearn.metrics import accuracy_score
from classification_model.config.core import config


def test_make_prediction(sample_input_data):
    y_true = sample_input_data[config.model_config.target]
    sample_input_data.drop(columns=[config.model_config.target],inplace=True)
    print(sample_input_data.columns)
    # Given
    expected_no_predictions = 262

    # When
    result = make_prediction(data_input=sample_input_data)
    print(result.get("errors"))

    # Then
    predictions = result.get("predictions")
    assert isinstance(predictions, np.ndarray)
    assert isinstance(predictions[0], np.int64)
    assert result.get("errors") is None
    assert len(predictions) == expected_no_predictions
    _predictions = list(predictions)
    accuracy = accuracy_score(_predictions, y_true)
    assert accuracy > 0.6



from feature_engine.encoding import OneHotEncoder
from feature_engine.imputation import AddMissingIndicator, CategoricalImputer, MeanMedianImputer
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline

# feature scaling
from sklearn.preprocessing import StandardScaler



from classification_model.config.core import config
from classification_model.processing import features as pp

# set up the pipeline
titanic_pipe = Pipeline([

    # ===== IMPUTATION =====
    # impute categorical variables with string missing
    ('categorical_imputation', CategoricalImputer(
        imputation_method='missing', variables=config.model_config.CATEGORICAL_VARIABLES
    )
    ),

    # add missing indicator to numerical variables
    ('missing_indicator', AddMissingIndicator(variables=config.model_config.NUMERICAL_VARIABLES)),

    # impute numerical variables with the median
    ('median_imputation', MeanMedianImputer(
        imputation_method='median', variables=config.model_config.NUMERICAL_VARIABLES)),


    # Extract letter from cabin
    ('extract_letter', pp.ExtractLetterTransformer(variable=config.model_config.CABIN)),
    
    # == CATEGORICAL ENCODING ======
    # remove categories present in less than 5% of the observations (0.05)
    # group them in one category called 'Rare'
    ('rare_label_encoder', pp.RareLabelCategoricalEncoder(variables=config.model_config.CATEGORICAL_VARIABLES,
                                         rare_perc=config.model_config.rare_perc)),

    # encode categorical variables using one hot encoding into k-1 variables
    ('categorical_encoder', OneHotEncoder(
        drop_last=True, variables=config.model_config.CATEGORICAL_VARIABLES)),

    # scale
    ('scaler', StandardScaler()),
    ('Logit', LogisticRegression(C=config.model_config.C, random_state=config.model_config.random_state)
),
]
)

